package de.unihannover.orderingExperimentAnalysis;

import java.io.IOException;
import java.io.StringReader;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashSet;
import java.util.LinkedHashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Map.Entry;
import java.util.Set;
import java.util.TreeMap;
import java.util.function.Function;
import java.util.stream.Collectors;

import de.unihannover.orderingExperimentUi.DiffData;
import de.unihannover.orderingExperimentUi.Treatment;
import de.unihannover.orderingExperimentUi.DiffData.ChangePart;
import de.unihannover.orderingExperimentUi.DiffData.FragmentData;

public class FindRelations {

    public static void main(String[] args) throws IOException {
        analyzeRelations("c2");
    }

    private static void analyzeRelations(String change) throws IOException {
        final List<ChangePart> parts = DiffData.load(change, Treatment.OPTIMAL_FILES).getChangeParts();

        System.out.println("## Files");
        printGroups(parts, (ChangePart p) -> Collections.singleton(p.getRight().getHeader()));
        System.out.println();

        System.out.println("## Overrides");
        printGroups(parts, FindRelations::getOverrides);
        System.out.println();

        System.out.println("## Attribute declaration and use");
        final Map<String, Set<String>> atts = determineDeclaredAttributes(parts);
        printAttributeUses(parts, atts);
        System.out.println();

        System.out.println("## Common tokens");
        printGroups(parts, FindRelations::getTokensNew);
        System.out.println();

        System.out.println("## Highly related (new)");
        printHighlyRelated(parts);
        System.out.println();

        System.out.println("## Probably moved code");
        printPotentialMoves(parts);
        System.out.println();
    }

    private static void printPotentialMoves(List<ChangePart> parts) {
        for (int i = 0; i < parts.size(); i++) {
            for (int j = 0; j < parts.size(); j++) {
                if (i == j) {
                    continue;
                }
                if (!(parts.get(i).getRight().getContent().isEmpty() && parts.get(j).getLeft().getContent().trim().isEmpty())) {
                    //we only look at moves where code was added somewhere and removed somewhere else
                    continue;
                }
                final FragmentData left = parts.get(i).getLeft();
                final FragmentData right = parts.get(j).getRight();
                final Set<String> t1 = getTokens(left.getContent());
                final Set<String> t2 = getTokens(right.getContent());
                final Set<String> intersect = new HashSet<>(t1);
                intersect.retainAll(t2);
                if (intersect.size() < 10) {
                    continue;
                }
                final double dist = (intersect.size() + 1.0) / (Math.min(t1.size(), t2.size()) + 1.0);
                if (dist > 0.8) {
                    System.out.println(parts.get(i).getId() + "," + parts.get(j).getId() + ": " + dist);
                }
            }
        }
    }

    private static void printHighlyRelated(List<ChangePart> parts) {
        for (int i = 0; i < parts.size(); i++) {
            for (int j = i + 1; j < parts.size(); j++) {
                final double dist = determineSimilarity(
                                parts.get(i).getRight().getContent(),
                                parts.get(j).getRight().getContent());
                if (dist > 0.7) {
                    System.out.println(parts.get(i).getId() + "," + parts.get(j).getId() + ": " + dist);
                }
            }
        }
    }

    private static double determineSimilarity(String c1, String c2) {
        final List<String> t1 = tokenize(c1);
        final List<String> t2 = tokenize(c2);
        final Set<String> union = new HashSet<>(t1);
        final Set<String> intersect = new HashSet<>(t1);
        union.addAll(t2);
        intersect.retainAll(t2);
        final double dist = ((double) intersect.size()) / union.size();
        return dist;
    }

    private static void printGroups(List<ChangePart> parts, Function<ChangePart, Set<String>> classifier) {
        final Map<String, List<ChangePart>> grouped = new TreeMap<>();
        for (final ChangePart p : parts) {
            System.out.println(p.getId() + ": " + classifier.apply(p));
            for (final String key : classifier.apply(p)) {
                List<ChangePart> list = grouped.get(key);
                if (list == null) {
                    list = new ArrayList<>();
                    grouped.put(key, list);
                }
                list.add(p);
            }
        }

        for (final Entry<String, List<ChangePart>> e : grouped.entrySet()) {
            if (e.getValue().size() > 1) {
                System.out.println(e.getKey() + ":");
                System.out.println("  " + e.getValue().stream().map(ChangePart::getId).collect(Collectors.toList()));
            }
        }
    }

    private static Set<String> getOverrides(ChangePart cp) {
        final Set<String> ret = new LinkedHashSet<>();
        final String methodNameNew = findMethodName(cp.getPrefix() + "\n" + cp.getRight().getContent());
        if (methodNameNew != null) {
            ret.add("new:" + methodNameNew);
        }
        final String methodNameOld = findMethodName(cp.getPrefix() + "\n" + cp.getLeft().getContent());
        if (methodNameOld != null) {
            ret.add("old:" + methodNameOld);
        }
        return ret;
    }

    private enum FindMethodState {
        OUT,
        HAD_BRACE,
        HAD_PAREN_CLOSE,
        HAD_PAREN_OPEN,
        NAME_FOUND
    }

    private static String findMethodName(String codeFragment) {
        final List<String> tokens = tokenize(codeFragment);
        if (tokens.isEmpty()) {
            return null;
        }
        Collections.reverse(tokens);
        if (tokens.get(0).equals("}")) {
            //if a whole method was added, position the scanner inside of it
            tokens.remove(0);
        }

        int nestingDepth = 0;
        String name = null;
        FindMethodState state = FindMethodState.OUT;
        for (final String token : tokens) {
            switch(state) {
            case OUT:
                if (token.equals("{")) {
                    if (nestingDepth == 0) {
                        state = FindMethodState.HAD_BRACE;
                    }
                    if (nestingDepth > 0) {
                        nestingDepth--;
                    }
                } else if (token.equals("}")) {
                    nestingDepth++;
                }
                break;
            case HAD_BRACE:
                if (token.equals(")")) {
                    state = FindMethodState.HAD_PAREN_CLOSE;
                } else {
                    state = FindMethodState.OUT;
                }
                break;
            case HAD_PAREN_CLOSE:
                if (token.equals("(")) {
                    state = FindMethodState.HAD_PAREN_OPEN;
                }
                break;
            case HAD_PAREN_OPEN:
                if (token.equals("if") || token.equals("while") || token.equals("for")) {
                    state = FindMethodState.OUT;
                } else if (isPotentialMethodName(token)) {
                    name = token;
                    state = FindMethodState.NAME_FOUND;
                } else {
                    state = FindMethodState.OUT;
                }
                break;
            case NAME_FOUND:
                if (token.equals(";") || token.equals("{") || token.equals("}")) {
                    return name;
                } else if (token.equals("class") || token.equals("interface") || token.equals("enum") || token.equals("new")) {
                    state = FindMethodState.OUT;
                }
                break;
            }
        }
        if (state == FindMethodState.NAME_FOUND) {
            return name;
        }
        return null;
    }

    private static boolean isPotentialMethodName(String token) {
        for (int i = 0; i < token.length(); i++) {
            if (!Character.isJavaIdentifierPart(token.charAt(i))) {
                return false;
            }
        }
        return true;
    }

    private static List<String> tokenize(String codeFragment) {
        final List<String> tokens = new ArrayList<>();
        final Tokenizer t = new Tokenizer(new StringReader(codeFragment));
        String token;
        try {
            while ((token = t.next()) != null) {
                tokens.add(token);
            }
        } catch (final IOException e) {
            throw new AssertionError(e);
        }
        return tokens;
    }

    private static Map<String, Set<String>> determineDeclaredAttributes(List<ChangePart> parts) {
        final TreeMap<String, Set<String>> map = new TreeMap<>();
        for (final ChangePart p : parts) {
            final List<String> cp = new ArrayList<>();
            final List<String> cs = new ArrayList<>();
            determineCommonPrefixAndSuffix(p.getLeft(), p.getRight(), cp, cs);
            map.put("new:" + p.getId(), getDeclaredAttributes(p.getRight().shrink(cp.size(), cs.size()).getContent()));
            map.put("old:" + p.getId(), getDeclaredAttributes(p.getLeft().shrink(cp.size(), cs.size()).getContent()));
        }
        return map;
    }

    private static void determineCommonPrefixAndSuffix(
                    FragmentData left,
                    FragmentData right,
                    List<String> commonPrefixBuffer,
                    List<String> commonSuffixBuffer) {
        final LinkedList<String> linesLeft = new LinkedList<>(left.getLines());
        final LinkedList<String> linesRight = new LinkedList<>(right.getLines());
        boolean didSomething;
        do {
            didSomething = false;
            if (linesLeft.isEmpty() || linesRight.isEmpty()) {
                break;
            }
            if (linesLeft.getFirst().equals(linesRight.getFirst())) {
                commonPrefixBuffer.add(linesLeft.removeFirst());
                linesRight.removeFirst();
                didSomething = true;
            }
            if (linesLeft.isEmpty() || linesRight.isEmpty()) {
                break;
            }
            if (linesLeft.getLast().equals(linesRight.getLast())) {
                commonSuffixBuffer.add(linesLeft.removeLast());
                linesRight.removeLast();
                didSomething = true;
            }
        } while (didSomething);
    }

    private static Set<String> getDeclaredAttributes(String changedCode) {
        final List<String> tokens = tokenize(changedCode);
        boolean canBeAttr = false;
        String prevToken = null;
        final Set<String> atts = new LinkedHashSet<>();
        for (final String token : tokens) {
            if (token.equals("private") || token.equals("protected") || token.equals("public")) {
                canBeAttr = true;
            } else if (token.equals(";") || token.equals("=")) {
                if (canBeAttr && prevToken != null) {
                    atts.add(prevToken);
                }
                canBeAttr = false;
            } else if (token.equals("(") || token.equals("{")) {
                canBeAttr = false;
            }
            prevToken = token;
        }
        return atts;
    }

    private static void printAttributeUses(List<ChangePart> parts, Map<String, Set<String>> atts) {
        for (final Entry<String, Set<String>> e : atts.entrySet()) {
            for (final String attName : e.getValue()) {
                final Set<String> matches = new LinkedHashSet<>();
                matches.add(e.getKey().substring(4));
                for (final ChangePart part : parts) {
                    Set<String> tokens;
                    if (e.getKey().startsWith("new")) {
                        tokens = getTokens(part.getRight().getContent());
                    } else {
                        tokens = getTokens(part.getLeft().getContent());
                    }
                    if (tokens.contains(attName)) {
                        matches.add(part.getId());
                    }
                }
                if (matches.size() > 1) {
                    System.out.println(e.getKey().substring(0, 4) + attName);
                    System.out.println("  " + matches);
                }
            }
        }
    }

    private static Set<String> getTokensNew(ChangePart part) {
        return getTokens(part.getRight().getContent());
    }

    private static Set<String> getTokens(String content) {
        return new HashSet<>(tokenize(content));
    }

}
